from web3 import Web3

from dydx3.constants import OFF_CHAIN_ONBOARDING_ACTION
from dydx3.constants import OFF_CHAIN_KEY_DERIVATION_ACTION
from dydx3.eth_signing import SignOnboardingAction
from dydx3.helpers.requests import request


class Onboarding(object):

    def __init__(
        self,
        host,
        eth_signer,
        network_id,
        default_address,
        stark_public_key=None,
        stark_public_key_y_coordinate=None,
    ):
        self.host = host
        self.default_address = default_address
        self.stark_public_key = stark_public_key
        self.stark_public_key_y_coordinate = stark_public_key_y_coordinate

        self.signer = SignOnboardingAction(eth_signer, network_id)

    # ============ Request Helpers ============

    def _post(
        self,
        endpoint,
        data,
        opt_ethereum_address,
    ):
        ethereum_address = opt_ethereum_address or self.default_address

        signature = self.signer.sign(
            ethereum_address,
            action=OFF_CHAIN_ONBOARDING_ACTION,
        )

        request_path = '/'.join(['/v3', endpoint])
        return request(
            self.host + request_path,
            'post',
            {
                'DYDX-SIGNATURE': signature,
                'DYDX-ETHEREUM-ADDRESS': ethereum_address,
            },
            data,
        )

    # ============ Requests ============

    def create_user(
        self,
        stark_public_key=None,
        stark_public_key_y_coordinate=None,
        ethereum_address=None,
    ):
        '''
        Onboard a user with an Ethereum address and STARK key.

        By default, onboards using the STARK and/or API public keys
        corresponding to private keys that the client was initialized with.

        :param stark_public_key: optional
        :type stark_public_key: str

        :param stark_public_key_y_coordinate: optional
        :type stark_public_key_y_coordinate: str

        :param ethereum_address: optional
        :type ethereum_address: str

        :returns: { apiKey, user, account }

        :raises: DydxAPIError
        '''
        stark_key = stark_public_key or self.stark_public_key
        stark_key_y = (
            stark_public_key_y_coordinate or self.stark_public_key_y_coordinate
        )
        if stark_key is None:
            raise ValueError(
                'STARK private key or public key is required'
            )
        if stark_key is None:
            raise ValueError(
                'STARK private key or public key y-coordinate is required'
            )
        return self._post(
            'onboarding',
            {
                'starkKey': stark_key,
                'starkKeyYCoordinate': stark_key_y,
            },
            ethereum_address,
        )

    # ============ Key Derivation ============

    def derive_stark_key(
        self,
        ethereum_address,
    ):
        '''
        Derive a STARK key pair deterministically from an Ethereum key.

        This is the function used by the dYdX frontend to derive a user's
        STARK key pair in a way that is recoverable. Programmatic traders may
        optionally derive their STARK key pair in the same way.

        :param ethereum_address: optional
        :type ethereum_address: str
        '''
        signature = self.signer.sign(
            ethereum_address or self.default_address,
            action=OFF_CHAIN_KEY_DERIVATION_ACTION,
        )
        signature_int = int(signature, 16)
        hashed_signature = Web3.solidityKeccak(['uint256'], [signature_int])
        private_key_int = int(hashed_signature.hex(), 16) >> 5
        return hex(private_key_int)



from dydx3.starkex.api_request import SignableApiRequest

MOCK_PUBLIC_KEY = (
    '3b865a18323b8d147a12c556bfb1d502516c325b1477a23ba6c77af31f020fd'
)
MOCK_PRIVATE_KEY = (
    '58c7d5a90b1776bde86ebac077e053ed85b0f7164f53b080304a531947f46e3'
)
MOCK_SIGNATURE = (
    '05191fb93fc14a948a4185ccd4b78a3ef423426da2362f0bb9adb91977be1a9a' +
    '050d8fbaa17b429a70df6d4c938058e848b96496a00c8878f8e7a7f88037240b'
)


class TestApiRequest():

    def test_sign_api_request(self):
        api_request = SignableApiRequest(
            iso_timestamp='2020-10-19T20:31:20.000Z',
            method='GET',
            body='',
            request_path='/v3/users',
        )
        signature = api_request.sign(MOCK_PRIVATE_KEY)
        assert signature == MOCK_SIGNATURE



from collections import namedtuple

from dydx3.starkex.helpers import message_to_hash
from dydx3.starkex.signable import Signable

StarkwareApiRequest = namedtuple(
    'StarkwareApiRequest',
    [
        'iso_timestamp',
        'method',
        'request_path',
        'body',
    ],
)


class SignableApiRequest(Signable):

    def __init__(
        self,
        iso_timestamp,
        method,
        request_path,
        body,
    ):
        message = StarkwareApiRequest(
            iso_timestamp=iso_timestamp,
            method=method,
            request_path=request_path,
            body=body,
        )
        super(SignableApiRequest, self).__init__(message)

    @property
    def message(self):
        return self.message

    def _calculate_hash(self):
        """Calculate the hash of the Starkware ApiRequest."""

        message_string = (
            self._message.iso_timestamp
            + self._message.method
            + self._message.request_path
            + self._message.body
        )
        return message_to_hash(message_string)



from web3 import Web3

from dydx3 import Client

from tests.constants import DEFAULT_HOST


class TestOnboarding():

    def test_derive_stark_key(self):
        web3 = Web3()  # Connect to a local Ethereum node.
        client = Client(
            host=DEFAULT_HOST,
            web3=web3,
        )
        signer_address = web3.eth.accounts[0]
        stark_private_key = client.onboarding.derive_stark_key(signer_address)
        assert stark_private_key == (
            '0x1bb0389af265c56844fa79b1e586b36f535fee293f59768fdb61d3f280f05fb'
        )



from web3 import Web3

from dydx3.eth_signing import util
from dydx3.eth_signing.sign_off_chain_action import SignOffChainAction

EIP712_OFF_CHAIN_ACTION_STRUCT_STRING = (
    'dYdX(' +
    'string action' +
    ')'
)


class SignOnboardingAction(SignOffChainAction):

    def get_hash(
        self,
        action,
    ):
        data = [
            [
                'bytes32',
                'bytes32',
            ],
            [
                util.hash_string(EIP712_OFF_CHAIN_ACTION_STRUCT_STRING),
                util.hash_string(action),
            ],
        ]
        struct_hash = Web3.solidityKeccak(*data)
        return self.get_eip712_hash(struct_hash)
